#
# This script shows the usage of the libpyRR (rrLib) module.
# - It connects to the rrServer
# - Displays all clients with their CPU usage
# - Displays all jobs (or jobs of user X)
# - Displays all users set in rrConfig/rrLogins
#
#

import sys
import os

mod_dir = os.path.dirname(__file__)
if mod_dir not in sys.path:
    sys.path.append(mod_dir)

from rr_python_utils.load_rrlib import rrLib  # module libpyRR3 for python 3, libpyRR2 for python 2
import rr_python_utils.connection as rr_connect
import rrJob

print("rrLib version: "+rrLib.getRRModuleVersion())




if sys.version_info.major < 3:
    RANGE_FUNC = xrange
else:
    RANGE_FUNC = range


class InfoLevel:
    """There are three levels of job information:
        MinInfo: The list, all jobs with ID, user, software, project
        Basic:   Scene information
        Send:    All job data you can have
    """
    MinInfo = 0
    Basic = 1
    Send = 2


def print_clients(tcp):
    print("\nCheck clients")
    if not tcp.clientGetList():
        print("Error getting clients: " + tcp.errorMessage())
        return

    clients = tcp.clients
    clients_count = clients.count()
    print("Number of client found: " + str(clients_count))

    for i in RANGE_FUNC(clients_count):
        client = clients.at(i)
        print("\tCpuUsage {0:6.2f} name: {1}".format(client.CPU_Usage, client.name))


def print_jobs(tcp, job_user_name="", info_level=InfoLevel.Basic):
    ### --------------------------------------------------------------------- JOBS
    ### IMPORTANT: tcp caches all jobs, if you destroy it, the rrServer has to send all jobs again
    ### A list request will always be executed, so you have a list of all jobs at the rrServer
    ### But to reduce traffic, you should set a filter for the jobs you need before you request more data

    print("\nCheck jobs")
    # retrieve only job details of one user to reduce network traffic, rrServer workload and this Pythons memory.
    if rrLib.getModuleVersionMajor()<9:
        tcp.jobList_SetFilter3(job_user_name, "", "", 0, "", False, rrLib._filterIDs.isAll)
    else:
        tcp.jobList_SetFilter4(job_user_name, "", "", rrJob._renderType.stUnset, 0, "", False, rrLib._filterIDs.isAll)

    if info_level == InfoLevel.Basic:
        # we do not need the full information about a job
        if not tcp.jobList_GetBasic():
            print("Error getting jobs: " + tcp.errorMessage())
            return
    elif info_level == InfoLevel.Info:
        if not tcp.jobList_GetInfo():
            print("Error getting jobs: " + tcp.errorMessage())
            return

    jobs = tcp.jobs
    print("Number of jobs: " + str(jobs.getMaxJobs()))
    jobs_count = jobs.getMaxJobsFiltered()
    print("Number of jobs - filtered by user {0}: {1}".format(job_user_name, jobs_count))
    for i in RANGE_FUNC(0, jobs_count):
        job_id = jobs.getJobMinInfo_filterQueue(i).ID
        username = jobs.getJobMinInfo_filterQueue(i).userName
        job_data_type = jobs.getJobDataType(job_id)  # which kind of data we have for the job
        job = jobs.getJobBasic(job_id)
        job_line = "\t{0} -\tuser: {1}\tscene: {2}\timage dim: {3}x{4}".format(
            job.IDstr(), username, job.sceneName, job.imageWidth, job.imageHeight
        )
        print(job_line)


def change_jobs_sequence(tcp, jobs, n):
    """Change in/out Sequence of n-th job in list"""
    import random
    random.seed()

    modjob_values = rrJob.getClass_SettingsOnly()
    modjob_values.seqStart=random.randint(2,50)
    modjob_values.seqEnd=random.randint(80,120)

    modjob_flags = rrJob.getClass_SettingsOnly()
    modjob_flags.seqStart=1
    modjob_flags.seqEnd=1

    modjobList = []

    job_min_info = jobs.getJobMinInfo_filterQueue(n)
    job_id = job_min_info.ID
    modjobList.append(job_id)

    print("Changing job {0} sequence to min={1}  max={2}".format(job_min_info.IDstr(), modjob_values.seqStart, modjob_values.seqEnd))
    tcp.jobModify(modjobList, modjob_values, modjob_flags)


def print_users(tcp):
    print("\nCheck users")
    if not tcp.userGet("X"):  # need a non-zero string and returns all the users!!
        print("Error getting user: " + tcp.errorMessage())
        return

    users = tcp.users
    users_count = users.count()
    print("Number of user found: " + str(users_count))
    for i in RANGE_FUNC(0, users_count):
        print("\tname: " + users.at(i).name)


tcp_ = rr_connect.server_connect(user_name=None)
print_clients(tcp_)
print_jobs(tcp_)
print_users(tcp_)

print("\n\nConnection Stats " + tcp_.connectionStats())
print("\n\n--- DONE ---\n\n")
